# Copyright (c) 2010, Yaco Sistemas <lgs@yaco.es>.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE.

from repoze.what.plugins.hgwebdir.predicates import has_repo_access

MEDIA_FILES = ("static", "favicon.ico")


def _get_repo_from_path_info(path_info):
    parts = path_info.strip('/').split('/')
    if not parts:
        return None
    else:
        repo = parts[0]
        if repo and repo not in MEDIA_FILES:
            return repo


def protect_repositories(app, public_repos):

    def protected_app(environ, start_response):
        repo = _get_repo_from_path_info(environ['PATH_INFO'])
        if not repo or has_repo_access(repo, public_repos).is_met(environ):
            return app(environ, start_response)
        else:
            start_response('401 not authorized', [])
            return []

    return protected_app



# Copyright (c) 2010, Yaco Sistemas <lgs@yaco.es>.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE.

import os
import tempfile
import unittest

from repoze.what.adapters.testutil import ReadOnlyGroupsAdapterTester
from repoze.what.adapters.testutil import ReadOnlyPermissionsAdapterTester

from repoze.what.plugins.hgwebdir.adapters import HgwebdirGroupsAdapter
from repoze.what.plugins.hgwebdir.adapters import HgwebdirPermissionsAdapter
from repoze.what.plugins.hgwebdir.adapters import get_public_repositories
from repoze.what.plugins.hgwebdir.middleware import _get_repo_from_path_info
from repoze.what.plugins.hgwebdir.middleware import protect_repositories
from repoze.what.plugins.hgwebdir.predicates import is_read_access
from repoze.what.plugins.hgwebdir.predicates import has_repo_access


sample_authz = """
[groups]
rolling = mick, keith, ronnie, charlie
beatles = john, paul, george, ringo

[sticky-fingers]
@rolling = rw
@beatles = r

[sargent-peppers]
@beatles = rw
@rolling = r

[pet-sounds]
* = r

[white-album]
@beatles = rw

[imagine]
john = rw
* = r
"""


class CommonTester(unittest.TestCase):

    def setUp(self):
        super(CommonTester, self).setUp()
        self.authz_file = tempfile.mkstemp('repoze_what_plugins_hgwebdir')[1]
        self.write(sample_authz)

    def tearDown(self):
        super(CommonTester, self).tearDown()
        os.remove(self.authz_file)

    def write(self, data):
        fd = open(self.authz_file, 'w')
        fd.write(data)
        fd.close()


class HgwebdirGroupsTester(CommonTester, ReadOnlyGroupsAdapterTester):

    def setUp(self):
        super(HgwebdirGroupsTester, self).setUp()

        self.adapter = HgwebdirGroupsAdapter(self.authz_file)

        self.all_sections = {
            u'rolling': set((u'mick', u'keith', u'ronnie', u'charlie')),
            u'beatles': set((u'john', u'paul', u'george', u'ringo')),
            u'john': set((u'john', )),
            }


class HgwebdirPermissionsTester(CommonTester, ReadOnlyPermissionsAdapterTester):

    def setUp(self):
        super(HgwebdirPermissionsTester, self).setUp()

        self.adapter = HgwebdirPermissionsAdapter(self.authz_file)

        self.all_sections = {
            u'sticky-fingers-read': set((u'rolling', u'beatles')),
            u'sticky-fingers-write': set((u'rolling', )),
            u'sargent-peppers-read': set((u'rolling', u'beatles')),
            u'sargent-peppers-write': set((u'beatles', )),
            u'pet-sounds-read': set(),
            u'pet-sounds-write': set(),
            u'white-album-read': set((u'beatles', )),
            u'white-album-write': set((u'beatles', )),
            u'imagine-read': set((u'john', )),
            u'imagine-write': set((u'john', )),
            }


class PublicRepositoriesTester(CommonTester):

    def test_get_public_repositories(self):
        self.assertEquals(get_public_repositories(self.authz_file),
                          [u'pet-sounds', u'imagine'])


class PredicatesTester(unittest.TestCase):

    def test_read_access(self):
        assert is_read_access().is_met({'REQUEST_METHOD': 'GET'}), \
            "A GET should be read access"
        assert not is_read_access().is_met({'REQUEST_METHOD': 'POST'}), \
            "A POST should not be read access"
        assert not is_read_access().is_met({'REQUEST_METHOD': 'PUT'}), \
            "A PUT should not be read access"
        assert not is_read_access().is_met({'REQUEST_METHOD': 'DELETE'}), \
            "A DELETE should not be read access"

    def test_repo_access(self):
        environ = {
            'REQUEST_METHOD': 'GET',
            'repoze.what.credentials': {
                'repoze.what.userid': 'john',
                'groups': tuple(),
                'permissions': (u'repo1-read', u'repo1-write', u'repo2-read'),
                },
            }
        assert has_repo_access('repo1', []).is_met(environ)
        assert has_repo_access('repo2', []).is_met(environ)
        assert not has_repo_access('repo3', []).is_met(environ)

        environ['REQUEST_METHOD'] = 'POST'
        assert has_repo_access('repo1', []).is_met(environ)
        assert not has_repo_access('repo2', []).is_met(environ)
        assert not has_repo_access('repo3', []).is_met(environ)

        # test public repositories
        environ['REQUEST_METHOD'] = 'GET'
        assert has_repo_access('repo4', ['repo4']).is_met(environ)


class MiddlewareTester(unittest.TestCase):

    def test_get_repo_from_path_info(self):
        self.assertEquals(_get_repo_from_path_info(''), None)
        self.assertEquals(_get_repo_from_path_info('/'), None)
        self.assertEquals(_get_repo_from_path_info('/repo1'), 'repo1')
        self.assertEquals(_get_repo_from_path_info('/repo1/'), 'repo1')
        self.assertEquals(_get_repo_from_path_info('/repo1/foo'), 'repo1')
        self.assertEquals(_get_repo_from_path_info('/repo1/foo/'), 'repo1')

        # media files are not repositories
        self.assertEquals(_get_repo_from_path_info('/favicon.ico'), None)
        self.assertEquals(_get_repo_from_path_info('/static'), None)
        self.assertEquals(_get_repo_from_path_info('/static/icon.png'), None)
        self.assertEquals(_get_repo_from_path_info('/static/styles.css'), None)

    def test_protect_repositories(self):

        content_type_header = ('Content-type', 'text/plain')

        def hello_world(environ, start_response):
            start_response('200 OK', [content_type_header])
            return ['Hello world\n']

        protected_app = protect_repositories(hello_world, [])

        log = []

        def start_response(status, headers):
            log.append((status, headers))

        env = {
            'REQUEST_METHOD': 'GET',
            'PATH_INFO': '/',
            }
        response = protected_app(env, start_response)
        self.assertEquals(response, ['Hello world\n'])
        self.assertEquals(log[-1], ('200 OK', [content_type_header]))

        env = {
            'REQUEST_METHOD': 'GET',
            'PATH_INFO': '/repo1',
            }
        response = protected_app(env, start_response)
        self.assertEquals(response, [])
        self.assertEquals(log[-1], ('401 not authorized', []))

        env = {
            'REQUEST_METHOD': 'GET',
            'PATH_INFO': '/repo1',
            'repoze.what.credentials': {
                'repoze.what.userid': 'john',
                'groups': tuple(),
                'permissions': (u'repo1-read', u'repo1-write'),
                },
            }
        response = protected_app(env, start_response)
        self.assertEquals(response, ['Hello world\n'])
        self.assertEquals(log[-1], ('200 OK', [content_type_header]))

        env = {
            'REQUEST_METHOD': 'POST',
            'PATH_INFO': '/repo1',
            'repoze.what.credentials': {
                'repoze.what.userid': 'john',
                'groups': tuple(),
                'permissions': (u'repo1-read', u'repo1-write'),
                },
            }
        response = protected_app(env, start_response)
        self.assertEquals(response, ['Hello world\n'])
        self.assertEquals(log[-1], ('200 OK', [content_type_header]))

        env = {
            'REQUEST_METHOD': 'GET',
            'PATH_INFO': '/repo2',
            'repoze.what.credentials': {
                'repoze.what.userid': 'john',
                'groups': tuple(),
                'permissions': (u'repo1-read', u'repo1-write'),
                },
            }
        response = protected_app(env, start_response)
        self.assertEquals(response, [])
        self.assertEquals(log[-1], ('401 not authorized', []))


def test_suite():
    return unittest.TestSuite((
        unittest.makeSuite(HgwebdirGroupsTester),
        unittest.makeSuite(HgwebdirPermissionsTester),
        unittest.makeSuite(PredicatesTester),
        unittest.makeSuite(MiddlewareTester),
        ))

if __name__ == '__main__':
    unittest.main(defaultTest='test_suite')



# Copyright (c) 2010, Yaco Sistemas <lgs@yaco.es>.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE.

import os
from setuptools import setup, find_packages

from repoze.what.plugins.hgwebdir import version


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


setup(
    name='repoze.what.plugins.hgwebdir',
    version=version,
    description='A simple repoze.what plugin useful for Mercurial hgwebdir',
    long_description='\n\n'.join([read('README.txt'), read('CHANGES.txt')]),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Security",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        ],
    keywords="authorization web application server wsgi repoze",
    author="Yaco Sistemas",
    author_email="lgs@yaco.es",
    namespace_packages=["repoze", "repoze.what", "repoze.what.plugins"],
    url="https://www.lorenzogil.com/projects/repoze.what.plugins.hgwebdir/",
    license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
    packages=find_packages(),
    zip_safe=False,
    install_requires=[
        'distribute',
        'repoze.what',
        ],
    )



# Copyright (c) 2010, Yaco Sistemas <lgs@yaco.es>.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE.

import os
import ConfigParser

from repoze.what.adapters import BaseSourceAdapter, SourceError


def U(text, encoding='utf8'):
    """Simple function to convert a string into a unicode object"""
    return unicode(text, encoding)


class CommonAdapter(BaseSourceAdapter):
    """Abstract base class for the group and permission adapters"""

    def __init__(self, authz_file):
        """Base constructor.

        Reads the groups or permission information from the authz_file

        Arguments:

          authz_file: is a init style file with information about
                      groups and permissions
        """
        super(CommonAdapter, self).__init__(writable=False)

        if not os.path.exists(authz_file):
            raise SourceError('Unable to find the authorization file "%s"'
                              % authz_file)

        self._data = ConfigParser.ConfigParser()
        self._data.read(authz_file)

        self._sections = self._load_sections()

    def _load_sections(self):
        """Subclasses should implement this"""
        raise SourceError("This is implemented in the groups and "
                          "permissions adapter")

    def _find_sections(self, hint):
        """Subclasses should implement this"""
        raise SourceError("This is implemented in the groups and "
                          "permissions adapter")

    def _get_all_sections(self):
        """Return all sections in this source adapter"""
        return self._sections

    def _get_section_items(self, section):
        """Return the items of the given section"""
        return self._sections[section]

    def _item_is_included(self, section, item):
        """True if the given item is included in the given section"""
        return item in self._sections[section]

    def _section_exists(self, section):
        """True if the section is included in this source adapter"""
        return section in self._sections.keys()


class HgwebdirGroupsAdapter(CommonAdapter):
    """Source adapter for groups"""

    def _load_sections(self):
        """Construct the sections and items information of this source.

        The ini file with the information should have been read before.

        The groups can be found in two places:

          - Inside the [groups] section, every option is a group and its
            members are the comma/space separate items in the value.
          - Inside the other sections a line like 'user = rw' is handled
            by creating a group with just an item. The name of the group
            is the same as its only item.
        """
        sections = {}

        for section in self._data.sections():
            if section == 'groups':
                for group, items in self._data.items(section):
                    members = [U(it).strip(u",") for it in items.split()]
                    sections[U(group)] = set(members)
            else:
                for name in self._data.options(section):
                    if not name.startswith(u"@") and name != u'*':
                        group = U(name)
                        if group in sections.keys():
                            raise SourceError('There is already a group named '
                                              '"%s". Maybe you forgot an @'
                                              % group)
                        sections[group] = set((group, ))

        return sections

    def _find_sections(self, hint):
        """Return a set of groups (sections) the user belongs to.

        Arguments:
          hint: a repoze credentials dictionary. The userid is in the
                repoze.what.userid key
        """
        userid = hint['repoze.what.userid']
        answer = set()
        for section, items in self._sections.items():
            if userid in items:
                answer.add(section)
        return answer


class HgwebdirPermissionsAdapter(CommonAdapter):
    """Source adapter for permissions"""

    def _load_sections(self):
        """Construct the sections and items information of this source.

        The ini file with the information should have been read before.

        The permissions are built from the sections of the ini file.
        The 'groups' section is special and ignored. From all the other
        sections, two permissions are created, one for read access and
        one for write access. The groups that have these permissions
        are built from the options and values of such sections.
        """

        sections = {}

        for section in self._data.sections():
            if section == 'groups':
                continue

            read_permission = U(section) + u'-read'
            write_permission = U(section) + u'-write'
            sections[read_permission] = set()
            sections[write_permission] = set()

            for name, value in self._data.items(section):
                if name == u'*':
                    continue

                if name.startswith(u"@"):
                    group = U(name[1:])
                else:
                    group = U(name)

                if 'r' in value:
                    sections[read_permission].add(group)

                if 'w' in value:
                    sections[write_permission].add(group)

        return sections

    def _find_sections(self, hint):
        """Return the set of permission that has the group.

        Arguments:
          hint: the group name
        """
        groupid = hint
        answer = set()
        for section, items in self._sections.items():
            if groupid in items:
                answer.add(section)
        return answer


def get_public_repositories(authz_file):
    """Returns a list of repositories that should not be protected.

    These repositories have a special line inside their section:
      * = r
    that means that everybody (*) can read them.
    """
    if not os.path.exists(authz_file):
        raise ValueError('Unable to find the authorization file "%s"'
                         % authz_file)

    data = ConfigParser.ConfigParser()
    data.read(authz_file)

    public_repositories = []

    for section in data.sections():
        if section == 'groups':
            continue

        for name, value in data.items(section):
            if name == u'*' and value == u'r':
                public_repositories.append(U(section))
                break

    return public_repositories



# Copyright (c) 2010, Yaco Sistemas <lgs@yaco.es>.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE.

from repoze.what.predicates import All, Any, Not, Predicate
from repoze.what.predicates import has_permission


class is_read_access(Predicate):
    message = 'The method %(method)s does not use read access'

    def evaluate(self, environ, credentials):
        method = environ['REQUEST_METHOD']
        if method != 'GET':
            self.unmet(method=method)


class has_repo_access(Any):
    message = 'The user must have access to the repository %(repo_name)s'

    def __init__(self, repo_name, public_repos, **kwargs):
        self.repo_name = repo_name
        predicates = [
            All(is_read_access(), has_permission(repo_name + '-read')),
            All(Not(is_read_access()), has_permission(repo_name + '-write')),
            ]
        if repo_name in public_repos:
            predicates.append(is_read_access())

        super(has_repo_access, self).__init__(*predicates, **kwargs)

    def evaluate(self, environ, credentials):
        super(has_repo_access, self).evaluate(environ, credentials)
